/* --------------------------------------------------------

	6.836 RA2 Artificial Chem Simulator 2/20/2002
  
	FILE: acsim.c

	Author: edsinger@ai.mit, jessical@ai.mit.edu
	
	This file implements the AC in generic C. It should compile
	on any platform. 
		
		  		  			
-------------------------------------------------------------*/

#include "acsim.h"


/*--------------------- Globals --------------------------------*/

struct patch world[SIZE_WORLD][SIZE_WORLD];
struct bacteria all_bacteria[MAX_BACTERIA];
int num_bacteria;
int num_reactions;


/*------------------- Utility Functions -------------------------*/

int get_bacteria_count()
{
	return num_bacteria;
}
/*--------------------------------------------------------------*/

float get_bacteria_dir(int id)
{
	return all_bacteria[id].direction;
}
/*--------------------------------------------------------------*/

struct coord get_bacteria_loc(int id)
{
	return all_bacteria[id].loc;
}
/*--------------------------------------------------------------*/
float get_world_molecular_concentration(int mol, int x, int y)
{
	return world[y][x].vat[mol];
}
/*--------------------------------------------------------------*/


//return scalar that scales min/max between 0/1
float get_world_molecular_scale(int mol)
{
	int x,y;
	float min,max;
	min =1; max = -1;
	
	for (x =0;x<SIZE_WORLD;x++)
		for (y =0;y<SIZE_WORLD;y++)
		{
			if (world[y][x].vat[mol] < min)
				min = world[y][x].vat[mol];
			if (world[y][x].vat[mol] > max)
				max = world[y][x].vat[mol];
		}
		return 1.0/(max-min);
}
/*--------------------------------------------------------------*/
//return the number of molecules in the bacteria
float get_bacteria_vat_cnt(int id)
{
	int i;
	float cnt=0;

	for (i=0;i<NUM_MOLECULES;i++)
	{
		cnt+=all_bacteria[id].vat[i];
	}

	return cnt;
	
}
/*--------------------------------------------------------------*/

//conform locations to torroidal world
float grid(float val)
{
	if (val<0) val+=SIZE_WORLD;
	return fmod(val,SIZE_WORLD);
}

/*--------------------------------------------------------------*/

//return indices of moore neighborhood
struct neighborhood get_neighborhood(struct coord c)
{
	struct neighborhood nbh;

	if (c.x==0) c.x=SIZE_WORLD;
	if (c.y==0) c.y=SIZE_WORLD;
	
	nbh.idx[0].x = (int)grid(c.x + 1);
	nbh.idx[0].y = (int)grid(c.y);
	
	nbh.idx[1].x = (int)grid(c.x - 1);
	nbh.idx[1].y = (int)grid(c.y);
	
	nbh.idx[2].x = (int)grid(c.x);
	nbh.idx[2].y = (int)grid(c.y + 1);
	
	nbh.idx[3].x = (int)grid(c.x);
	nbh.idx[3].y = (int)grid(c.y - 1);
	
	nbh.idx[4].x = (int)grid(c.x + 1);
	nbh.idx[4].y = (int)grid(c.y + 1);
	
	nbh.idx[5].x = (int)grid(c.x - 1);
	nbh.idx[5].y = (int)grid(c.y + 1);
	
	nbh.idx[6].x = (int)grid(c.x + 1);
	nbh.idx[6].y = (int)grid(c.y - 1);
	
	nbh.idx[7].x = (int)grid(c.x - 1);
	nbh.idx[7].y = (int)grid(c.y - 1);
	
	
	return nbh;
}






/*--------------------   Simulator Core    -----------------------------*/





/*--------------------------------------------------------------*/

//Setup the world

#define BACTERIA_MOL_SAT	1000	//max number of molecules allowed in a bacteria vat
#define INIT_NUM_BACTERIA	1		//Initial bacteria population
#define INIT_NUM_FOOD		10		//Initial number of world patches with chemical(s)
#define INIT_FOOD_AMT		100		//Initial amount of chemical in each patch
#define NUM_REACTIONS		2		//Number of metabolic reactions


void create_world()
{
	int x,y,m,xx,yy,i,j;
	struct coord c;

	/* -------------------- Setup Reactions -----------------
		
		  A reaction rule is of the form {{2,2,0,0},{0,3,1,0} which would be
		  (2 MOLA + 2 MOLB) -> (3 MOLB + 1 MOLC) if MOLA = 0, MOLB=1,... 
		  
			It is up to you to verify that your rules conserve molecular weight.
			  
				
		*/
	
		
		//For each reaction, add an tuple to this struct
		//If you add molecules to the simulation, be sure and
		//increase the size of each tuple
		
		struct reaction rules[MAX_REACTIONS] = {
			{{1,0,0,0,0},{0,0,1,0,0}}, //1A->1C
			{{0,1,0,0,0},{0,0,0,1,0}} //1B->1D
		}; 

		
/* -------------------- Setup Bacteria ----------------- */
		
	
		
		//The initial amount of molecules in each bacteria's vat
		//If you add molecules to the simulation, increase the size
		//of the tuple
		float initial_concentrations[NUM_MOLECULES]={20,20,20,20,0};

		num_reactions = NUM_REACTIONS;
		num_bacteria=INIT_NUM_BACTERIA;


/* -------------------- Initialize world -----------------*/

	srand( (unsigned)time( NULL ) );

	for (x =0;x<SIZE_WORLD;x++)
		for (y =0;y<SIZE_WORLD;y++)
		{
			world[y][x].loc.x = x;
			world[y][x].loc.y = y;
			c.x=x;c.y=y;
			world[y][x].nbh = get_neighborhood(c);
			for (m =0;m<NUM_MOLECULES;m++)
				world[y][x].vat[m]=0;
		}
		
		/* -------------------- Setup Initial Chems in World  -----------------*/
		for (i=0;i<INIT_NUM_FOOD;i++)
		{
			xx=grid(SIZE_WORLD * (float)rand()/(float)RAND_MAX);
			yy=grid(SIZE_WORLD * (float)rand()/(float)RAND_MAX);
			world[yy][xx].vat[MOL_WORLD]=INIT_FOOD_AMT;
		}
		
		/* -------------------- Setup Initial Bacteria in World  -----------------*/
		
		for (i=0;i<num_bacteria;i++)
		{
			all_bacteria[i].direction = 2*M_PI * (float)rand()/(float)RAND_MAX; //random 
			all_bacteria[i].loc.x = floor(grid(SIZE_WORLD * (float)rand()/(float)RAND_MAX));
			all_bacteria[i].loc.y = floor(grid(SIZE_WORLD*(float)rand()/(float)RAND_MAX));
			for (j=0;j<num_reactions;j++)
				all_bacteria[i].rules[j]=rules[j];
			for (j=0;j<NUM_MOLECULES;j++)
				all_bacteria[i].vat[j] = initial_concentrations[j];
		}
		
}




/*--------------------------------------------------------------*/
/*
	Do one time step of chemical diffusion
	Each cell diffuses a percentage of its
	chemicals to its Moore neighborhood
	Conservation of mass is maintained */

#define DIFFUSION_RATE 0.1 

struct patch new_world[SIZE_WORLD][SIZE_WORLD];

void diffuse_world()
{
	int x,y,m,i;
	float val;
	struct coord nb;

	for (x =0;x<SIZE_WORLD;x++)
		for (y =0;y<SIZE_WORLD;y++)
		{
			for (m =0;m<NUM_MOLECULES;m++)
			{
				val=0;
				for (i=0;i<NEIGHBORHOOD_SIZE;i++)  //sum concentrations in neighborhood
				{
					nb = world[y][x].nbh.idx[i];
					val += world[y][x].vat[m] - world[(int)nb.y][(int)nb.x].vat[m];
				}
				val = (float)(DIFFUSION_RATE / NEIGHBORHOOD_SIZE)*val;
				new_world[y][x].vat[m]= val;
			}
		}
		//Now copy back over 
		for (x =0;x<SIZE_WORLD;x++)
			for (y =0;y<SIZE_WORLD;y++)
			{
				for (m =0;m<NUM_MOLECULES;m++)
					world[y][x].vat[m]= world[y][x].vat[m]-new_world[y][x].vat[m];
			}
			
			
}

/*--------------------------------------------------------------*/


//randomly evaporate off chemicals in the world
#define PROB_EVAP .01
void evaporate_world()
{
	int x,y,m;

	for (x =0;x<SIZE_WORLD;x++)
		for (y =0;y<SIZE_WORLD;y++)
			for (m =0;m<NUM_MOLECULES;m++)
			{
				if ( (float)rand()/ (float)RAND_MAX < PROB_EVAP)
					world[y][x].vat[m] = max(0,world[y][x].vat[m]-1);
			}
			
}



/*--------------------------------------------------------------
	Sense and consume chemical(s) from the world here
	Add them to the internal vat up to BACTERIA_MOL_SAT
	In the spirit of the Ziegler paper, the bacteria can sense
	MOL_WORLD with 2 sensors. MOL_WORLD is converted to MOL_L_SENSE
	and MOL_R_SENSE in the internal metabolism, depending on which
	sensor it was sensed with. The bacteria consumes a percentage
	(INHALE_PCT) of the sensed chemicals from the cell it is sitting
	on. You can add more sensors or chemicals
	in the world if it makes life easier.  
	*/
#define INHALE_PCT 1.0
void bacteria_sensor(int id)
{
	
	struct coord cl, cr;
	float l,r,c,empty,pct_in;
	struct bacteria * b = &all_bacteria[id];

	//sense +/- 45 degrees
	cl.x = grid(b->loc.x + floor(cos(b->direction- M_PI/4.0)+ 0.5));
	cl.y = grid(b->loc.y + floor(sin(b->direction- M_PI/4.0)+ 0.5));
	cr.x = grid(b->loc.x + floor(cos(b->direction+ M_PI/4.0)+ 0.5));
	cr.y = grid(b->loc.y + floor(sin(b->direction+ M_PI/4.0)+ 0.5));
	
	empty = BACTERIA_MOL_SAT - get_bacteria_vat_cnt(id);
	l = INHALE_PCT * world[(int)cl.y][(int)cl.x].vat[MOL_WORLD];
	r = INHALE_PCT * world[(int)cr.y][(int)cr.x].vat[MOL_WORLD];
	c = world[(int)b->loc.y][(int)b->loc.x].vat[MOL_WORLD];
	if (l+r == 0) return;

	pct_in = min(empty / (l+r),1.0);  //make sure don't exceed max vat saturation
	pct_in = min(pct_in, c / (l+r));  //make sure don't exceed amt in patch we are on
	b->vat[MOL_L_SENSE] += pct_in*l;
	b->vat[MOL_R_SENSE] += pct_in*r;
	
	//Consume from world
	world[(int)b->loc.y][(int)b->loc.x].vat[MOL_WORLD]-=pct_in*l;
	world[(int)b->loc.y][(int)b->loc.x].vat[MOL_WORLD]-=pct_in*r;

	
}


/*--------------------------------------------------------------
 Bacteria locomotion currently is a product of MOL_L_THRUST and MOL_R_THRUST
 which is internally metabolized by the AC. The ratio of the two, scaled 
 by DIR_SCALE determines the directional adjustment. If there is sufficient 
 MOL_L_THRUST and MOL_R_THRUST, a change in direction and/or unit step is taken.

 This is a fairly unintersting method of locomotion on not very realistic. You may
 want to experiment with the method described in the Ziegler paper.
 */

#define STEP_ENERGY 70 
#define STEP_THRESH 70
#define DIR_ENERGY 1
#define DIR_THRESH 1
#define DIR_SCALE 0.05

void bacteria_locomotion(int id)
{
	float lt,rt,step_dir,plt,prt;

	struct bacteria * b = &all_bacteria[id];
	
	lt = b->vat[MOL_L_THRUST];
	rt = b->vat[MOL_R_THRUST];
	
	if (lt+rt == 0) return;

	plt = lt/(lt+rt); prt = rt/(lt+rt);
	
	if (lt+rt > DIR_THRESH)
	{
		//adjust direction proportional to ratio of left/right thrust
		step_dir =	b->direction - DIR_SCALE *(M_PI/2.0)*(rt/(lt+rt)) + DIR_SCALE *(M_PI/2.0)*(lt/(lt+rt));
		b->direction = fmod(step_dir,2*M_PI); 
		b->vat[MOL_L_THRUST] = max(0,b->vat[MOL_L_THRUST]- plt*DIR_ENERGY);
		b->vat[MOL_R_THRUST] = max(0,b->vat[MOL_R_THRUST]- prt*DIR_ENERGY);
	}
	
	
	
	//take a unit step 
	if (lt+rt >STEP_THRESH)
	{
		b->loc.x = grid(b->loc.x + floor(cos (b->direction) + 0.5));
		b->loc.y = grid(b->loc.y + floor(sin (b->direction) + 0.5));
		b->vat[MOL_L_THRUST] = max(0,b->vat[MOL_L_THRUST]-plt*STEP_ENERGY);
		b->vat[MOL_R_THRUST] = max(0,b->vat[MOL_R_THRUST]-prt*STEP_ENERGY);
	}
	
	
}

/*--------------------------------------------------------------
	The reactor uses a probablistic approach to determine which
	reaction rules to apply to the metabolism. The probability 
	of a particular rule is calculated. This is a function
	of the reaction length, and the number of molecules, and the
	concentration of that reactions molecules in the vat. 

	A random draw is used on the reaction rules proportional 
	to there probability. The rule is then applied by updating
	the concentrations in the vat.

    In some infrequent boundary cases, matter is not conserved. 
	You may want to revise the reactor to impose this constraint.

	REACTION_SIZE determines the number of times the selected rule
	is applied. 
*/
#define REACTION_SIZE 10
void bacteria_reactor(int id)
{
	int i,j,id_pick;
	float pick,add,remove;
	float sum=0;
	float rsum=0;
	float p[MAX_REACTIONS];  //prob of each reaction

	struct bacteria * b = &all_bacteria[id];
	
	sum = get_bacteria_vat_cnt(id);
	
	//calc prob of each reaction
	for (i=0;i<num_reactions;i++)
	{
		p[i]=1;
		for (j=0;j<NUM_MOLECULES;j++)
		{
			p[i] *= pow( (b->vat[j]/sum),(b->rules[i].lhs[j]));
		}
	}
	
	//norm to 1
	for (i=0;i<num_reactions;i++)
		rsum += p[i];
	
	if (rsum==0) //no prob of anything happening
		return;
	
	//random pick of reaction
	pick = (float)rand()/(float)RAND_MAX;
	id_pick=0;
	sum = 0;
	for (i=0;i<num_reactions;i++)
	{
		if (pick >= sum && pick <= sum+(p[i]/rsum))
			id_pick= i;
		sum +=(p[i]/rsum);
	}
	
	//update concentrations
	sum = get_bacteria_vat_cnt(id);
	for (i=0;i<NUM_MOLECULES;i++)
	{
		remove = b->rules[id_pick].lhs[i] * REACTION_SIZE;
		add = b->rules[id_pick].rhs[i] * REACTION_SIZE;
		
		if (remove > b->vat[i])  //hack to watch limits...no consv of matter
			remove = b->vat[i];

		b->vat[i] -= remove;

		if (add +sum > BACTERIA_MOL_SAT)
			add = max(0,BACTERIA_MOL_SAT - sum);

		b->vat[i] += add;
		sum+=add;
	}
	
	
}


